/*************************************************
* Simple ASN.1 String Types Source File          *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/asn1_obj.h>

namespace Botan {

namespace {

/*************************************************
* Choose an encoding for the string              *
*************************************************/
ASN1_Tag choose_encoding(const std::string& str)
   {
   byte mask = 0x01;

   for(u32bit j = 0; j != str.size(); j++)
      {
      while(!ASN1_String::valid_char(str[j], mask) && mask <= 0x80)
         mask <<= 1;
      }

   if(mask == 0x01)      return NUMERIC_STRING;
   else if(mask == 0x02) return PRINTABLE_STRING;
   else if(mask == 0x04) return VISIBLE_STRING;
   else if(mask == 0x08) return IA5_STRING;
   else if(mask == 0x10) return T61_STRING;
   throw Encoding_Error("No way to encode " + str + " as an ASN.1 string");
   }

}

/*************************************************
* Create an ASN1_String                          *
*************************************************/
ASN1_String::ASN1_String(const std::string& str_arg, ASN1_Tag t) :
   str(str_arg), tag(t == DIRECTORY_STRING ? choose_encoding(str_arg) : t)
   {
   if(tag != NUMERIC_STRING && tag != PRINTABLE_STRING &&
      tag != VISIBLE_STRING && tag != IA5_STRING && tag != T61_STRING)
      throw Invalid_Argument("ASN1_String: Invalid tag " + to_string(tag));

   byte mask = 0;
   if(tag == NUMERIC_STRING)        mask = 0x01;
   else if(tag == PRINTABLE_STRING) mask = 0x02;
   else if(tag == VISIBLE_STRING)   mask = 0x04;
   else if(tag == IA5_STRING)       mask = 0x08;
   else if(tag == T61_STRING)       mask = 0x10;

   for(u32bit j = 0; j != str_arg.size(); j++)
      if(!valid_char(str_arg[j], mask))
         throw Invalid_Argument("ASN1_String: String contains invalid chars");
   }

/*************************************************
* Create an ASN1_String                          *
*************************************************/
ASN1_String::ASN1_String(const std::string& str_arg) :
   str(str_arg), tag(choose_encoding(str))
   {
   }

/*************************************************
* Check the validity of input characters         *
*************************************************/
bool ASN1_String::valid_char(char c, byte mask)
   {
   const byte VALID_CODEPOINTS[256] = {
      0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08,
      0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08,
      0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x08, 0x1F, 0x1C, 0x1C, 0x1C,
      0x1C, 0x1C, 0x1C, 0x1E, 0x1E, 0x1E, 0x1C, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E,
      0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1E, 0x1C,
      0x1C, 0x1E, 0x1C, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E,
      0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E,
      0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C,
      0x1C, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E,
      0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E,
      0x1E, 0x1E, 0x1E, 0x1C, 0x1C, 0x1C, 0x1C, 0x18, 0x10, 0x10, 0x10, 0x10,
      0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
      0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
      0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
      0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
      0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
      0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
      0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
      0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
      0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
      0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10,
      0x10, 0x10, 0x10, 0x10 };
   byte point = (byte)c;
   return (VALID_CODEPOINTS[point] & mask);
   }

/*************************************************
* Return the value of the string                 *
*************************************************/
std::string ASN1_String::value() const
   {
   return str;
   }

/*************************************************
* Return the type of this string object          *
*************************************************/
ASN1_Tag ASN1_String::tagging() const
   {
   return tag;
   }

namespace DER {

/*************************************************
* DER encode an ASN1_String                      *
*************************************************/
void encode(DER_Encoder& encoder, const ASN1_String& string)
   {
   encoder.add_object(string.tagging(), UNIVERSAL, string.value());
   }

/*************************************************
* DER encode an ASN1_String, IMPLICIT tag        *
*************************************************/
void encode(DER_Encoder& encoder, const ASN1_String& string,
            ASN1_Tag type_tag, ASN1_Tag class_tag)
   {
   encoder.add_object(type_tag, class_tag, string.value());
   }

}

namespace BER {

/*************************************************
* Decode a BER encoded ASN1_String               *
*************************************************/
void decode(BER_Decoder& source, ASN1_String& string)
   {
   BER_Object ber_obj = source.get_next_object();
   string = ASN1_String(BER::to_string(ber_obj), ber_obj.tag);
   }

/*************************************************
* Decode a BER encoded ASN1_String               *
*************************************************/
void decode(BER_Decoder& source, ASN1_String& string,
            ASN1_Tag expected_tag, ASN1_Tag real_tag)
   {
   BER_Object ber_obj = source.get_next_object();
   if(ber_obj.tag != expected_tag)
      throw Decoding_Error("Unexpected tag in BER string decoding");
   string = ASN1_String(BER::to_string(ber_obj), real_tag);
   }

}

}
