/*************************************************
* Initalization Function Source File             *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/init.h>
#include <botan/allocate.h>
#include <botan/rng.h>
#include <botan/mutex.h>
#include <botan/timers.h>
#include <botan/look_add.h>
#include <botan/randpool.h>
#include <botan/x917_rng.h>
#include <botan/util.h>
#include <botan/conf.h>
#include <map>

#if defined(BOTAN_EXT_MUTEX_PTHREAD)
  #include <botan/mux_pthr.h>
#endif

#if defined(BOTAN_EXT_ALLOC_MMAP)
  #include <botan/mmap_mem.h>
#endif

#if defined(BOTAN_EXT_ALLOC_MLOCK)
  #include <botan/lock_mem.h>
#endif

#if defined(BOTAN_EXT_TIMER_HARDWARE)
  #include <botan/tm_hard.h>
#elif defined(BOTAN_EXT_TIMER_POSIX)
  #include <botan/tm_posix.h>
#elif defined(BOTAN_EXT_TIMER_UNIX)
  #include <botan/tm_unix.h>
#endif

namespace Botan {

/*************************************************
* Library Initalization                          *
*************************************************/
LibraryInitializer::LibraryInitializer(const std::string& arg_string)
   {
   Init::initialize(arg_string);
   }

/*************************************************
* Library Shutdown                               *
*************************************************/
LibraryInitializer::~LibraryInitializer()
   {
   Init::deinitialize();
   }

namespace {

/*************************************************
* Parse the option string                        *
*************************************************/
std::map<std::string, std::string> parse_args(const std::string& flags)
   {
   std::string opt_string = flags + " ";
   std::map<std::string, std::string> options;

   std::string::size_type start = 0, end = opt_string.find(" ");
   while(end != std::string::npos)
      {
      std::string option = opt_string.substr(start, end-start);
      start = end+1;
      end = opt_string.find(" ", start);

      if(option == "")
         continue;

      if(option.find("=") == std::string::npos)
         options[option] = "";
      else
         {
         std::string::size_type split = option.find("=");

         std::string opt_name = option.substr(0, split);
         std::string arg = option.substr(split + 1, option.length());

         if(opt_name == "")
            continue;

         options[opt_name] = arg;
         }
      }

   return options;
   }

/*************************************************
* Check if an option is set in the argument      *
*************************************************/
bool arg_set(const std::map<std::string, std::string>& args,
             const std::string& option)
   {
   if(args.find(option) != args.end())
      return true;
   return false;
   }

/*************************************************
* Register a mutex type, if possible             *
*************************************************/
bool set_mutex(bool must_have_mutex)
   {
   bool got_real_mutex = false;

#if defined(BOTAN_EXT_MUTEX_PTHREAD)
   set_mutex_type(new Pthread_Mutex);
   got_real_mutex = true;
#endif

   if(!must_have_mutex)
      return true;
   return got_real_mutex;
   }

/*************************************************
* Register a high resolution timer, if possible  *
*************************************************/
void set_timer()
   {
#if defined(BOTAN_EXT_TIMER_HARDWARE)
   set_timer_type(new Hardware_Timer);
#elif defined(BOTAN_EXT_TIMER_POSIX)
   set_timer_type(new POSIX_Timer);
#elif defined(BOTAN_EXT_TIMER_UNIX)
   set_timer_type(new Unix_Timer);
#endif
   }

/*************************************************
* Register a more secure allocator, if possible  *
*************************************************/
void set_safe_allocator()
   {
#if defined(BOTAN_EXT_ALLOC_MLOCK)
   add_allocator_type("mlock", new MemoryLocking_Allocator);
   set_default_allocator("mlock");
#endif

#if defined(BOTAN_EXT_ALLOC_MMAP)
   add_allocator_type("mmap", new MemoryMapping_Allocator);
   set_default_allocator("mmap");
#endif
   }

}

/*************************************************
* Internal startup/shutdown functions            *
*************************************************/
extern void initalize_memory_subsystem();
extern void shutdown_memory_subsystem();
extern void create_global_bigints();
extern void delete_global_bigints();
extern void init_lookup_tables();
extern void destroy_lookup_tables();
extern void load_algorithms_list();
extern void set_default_options();

namespace Init {

/*************************************************
* Library Initalization                          *
*************************************************/
void initialize(const std::string& arg_string)
   {
   std::map<std::string, std::string> args = parse_args(arg_string);

   if(!set_mutex(arg_set(args, "thread_safe")))
      throw Exception("LibraryInitializer: thread safety impossible");

   initalize_memory_subsystem();
   create_global_bigints();
   init_lookup_tables();
   load_algorithms_list();
   set_default_options();

   if(arg_set(args, "egd_path"))
      Config::add("rng/egd_path", args["egd_path"]);

   if(arg_set(args, "secure_memory"))
      set_safe_allocator();
   if(!arg_set(args, "no_timers"))
      set_timer();

   setup_global_rng(new Randpool, new ANSI_X917_RNG);

   Global_RNG::seed();
   }

/*************************************************
* Library Shutdown                               *
*************************************************/
void deinitialize()
   {
   destroy_global_rng();
   destroy_lookup_tables();
   delete_global_bigints();
   set_timer_type(0);
   shutdown_memory_subsystem();
   }

}

}
