/*************************************************
* PKCS #8 Source File                            *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/pkcs8.h>
#include <botan/asn1_obj.h>
#include <botan/pk_algs.h>
#include <botan/oids.h>
#include <botan/pem.h>
#include <botan/pbe.h>
#include <memory>

namespace Botan {

namespace PKCS8 {

namespace {

/*************************************************
* Extract the fields of a PrivateKeyInfo         *
*************************************************/
void PKCS8_extract_info(DataSource& source, AlgorithmIdentifier& alg_id,
                       SecureVector<byte>& pkcs8_key)
   {
   BigInt version;

   try {
      BER_Decoder decoder(source);
      BER_Decoder sequence = BER::get_subsequence(decoder);
      BER::decode(sequence, version);
      BER::decode(sequence, alg_id);
      BER::decode(sequence, pkcs8_key, OCTET_STRING);
      sequence.discard_remaining();
      sequence.verify_end();
      }
   catch(Decoding_Error)
      {
      throw Decoding_Error("PKCS #8 private key decoding failed");
      }

   if(version != 0)
      throw Decoding_Error("PKCS #8: Unknown version number detected");
   }

/*************************************************
* Decode a BER encoded PKCS #8 private key       *
*************************************************/
void PKCS8_BER_decode(PKCS8_PrivateKey& key, DataSource& source)
   {
   AlgorithmIdentifier alg_id;
   SecureVector<byte> pkcs8_key;

   PKCS8_extract_info(source, alg_id, pkcs8_key);

   if(!key.valid_oid(alg_id.oid))
      throw Decoding_Error("PKCS #8: OID mismatch, found " +
                           alg_id.oid.as_string() + ", expected " +
                           key.get_oid().as_string());

   Pipe output;
   output.process_msg(alg_id.parameters);
   output.process_msg(pkcs8_key);
   key.BER_decode_params(output);
   output.set_default_msg(1);
   key.BER_decode_priv(output);
   }

/*************************************************
* Decrypt an encrypted PKCS #8 private key       *
*************************************************/
SecureVector<byte> PKCS8_decrypt(DataSource& source, const std::string& pass)
   {
   AlgorithmIdentifier alg_id;
   SecureVector<byte> enc_pkcs8_key;

   try {
      BER_Decoder decoder(source);
      BER_Decoder sequence = BER::get_subsequence(decoder);
      BER::decode(sequence, alg_id);
      BER::decode(sequence, enc_pkcs8_key, OCTET_STRING);
      sequence.verify_end();

      Pipe params;
      params.process_msg(alg_id.parameters, alg_id.parameters.size());

      PBE* pbe = get_pbe(alg_id.oid.as_string(), params);

      pbe->set_key(pass);
      Pipe pkcs8_key(pbe);
      pkcs8_key.process_msg(enc_pkcs8_key, enc_pkcs8_key.size());
      return pkcs8_key.read_all();
   }
   catch(Decoding_Error)
      {
      throw Decoding_Error("PKCS #8 private key decoding failed");
      }
   }

/*************************************************
* PEM decode and/or decrypt a private key        *
*************************************************/
SecureVector<byte> PKCS8_PEM_handler(DataSource& source,
                                     const std::string& pass,
                                     X509_Encoding encoding)
   {
   if(encoding == PEM)
      {
      Pipe ber;
      std::string label;
      ber.process_msg(pem_decode(source, label));
      if(label == "PRIVATE KEY")
         return ber.read_all();
      else if(label == "ENCRYPTED PRIVATE KEY")
         return PKCS8_decrypt(ber, pass);
      throw Decoding_Error("PKCS #8: Unknown PEM label " + label);
      }
   return PKCS8_decrypt(source, pass);
   }

}

/*************************************************
* DER or PEM encode a PKCS #8 private key        *
*************************************************/
void encode(const PKCS8_PrivateKey& key, Pipe& pipe, X509_Encoding encoding)
   {
   DER_Encoder encoder;

   encoder.start_sequence();
   DER::encode(encoder, BigInt::zero());

   AlgorithmIdentifier alg_id(key.get_oid(), key.DER_encode_params());
   DER::encode(encoder, alg_id);

   DER::encode(encoder, key.DER_encode_priv(), OCTET_STRING);
   encoder.end_sequence();

   SecureVector<byte> der = encoder.get_contents();

   if(encoding == PEM)
      pipe.write(pem_encode(der, der.size(), "PRIVATE KEY"));
   else
      pipe.write(der);
   }

/*************************************************
* PEM encode a PKCS #8 private key               *
*************************************************/
std::string PEM_encode(const PKCS8_PrivateKey& key)
   {
   Pipe pem;
   pem.start_msg();
   encode(key, pem, PEM);
   pem.end_msg();
   return pem.read_all_as_string();
   }

/*************************************************
* Decode a PKCS #8 private key                   *
*************************************************/
void decode(PKCS8_PrivateKey& key, DataSource& source, X509_Encoding encoding)
   {
   if(encoding == PEM)
      {
      Pipe ber;
      ber.process_msg(pem_decode_check_label(source, "PRIVATE KEY"));
      PKCS8_BER_decode(key, ber);
      }
   else
      PKCS8_BER_decode(key, source);
   }

/*************************************************
* Encode and encrypt a PKCS #8 private key       *
*************************************************/
void encrypt_key(const PKCS8_PrivateKey& key, Pipe& pipe,
                 const std::string& pass, const std::string& pbe_algo,
                 X509_Encoding encoding)
   {
   Pipe raw_key;
   raw_key.start_msg();
   encode(key, raw_key, RAW_BER);
   raw_key.end_msg();

   PBE* pbe = get_pbe(pbe_algo);
   pbe->set_key(pass);
   AlgorithmIdentifier pbe_id(pbe->get_oid(), pbe->encode_params());
   Pipe encrypted_key(pbe);
   encrypted_key.process_msg(raw_key);

   DER_Encoder encoder;
   encoder.start_sequence();
   DER::encode(encoder, pbe_id);
   DER::encode(encoder, encrypted_key.read_all(), OCTET_STRING);
   encoder.end_sequence();
   SecureVector<byte> enc_key = encoder.get_contents();

   if(encoding == PEM)
      pipe.write(pem_encode(enc_key, enc_key.size(), "ENCRYPTED PRIVATE KEY"));
   else
      pipe.write(enc_key);
   }

/*************************************************
* Encrypt and PEM encode a PKCS #8 private key   *
*************************************************/
std::string PEM_encode(const PKCS8_PrivateKey& key, const std::string& pass)
   {
   Pipe pem;
   pem.start_msg();
   encrypt_key(key, pem, pass);
   pem.end_msg();
   return pem.read_all_as_string();
   }

/*************************************************
* Decrypt and decode a PKCS #8 private key       *
*************************************************/
void decrypt_key(PKCS8_PrivateKey& key, DataSource& in,
                 const std::string& pass, X509_Encoding encoding)
   {
   DataSource_Memory key_source(PKCS8_PEM_handler(in, pass, encoding));
   PKCS8_BER_decode(key, key_source);
   }

/*************************************************
* Extract a private key and return it            *
*************************************************/
PKCS8_PrivateKey* load_key(DataSource& in, const std::string& pass,
                           X509_Encoding encoding)
   {
   AlgorithmIdentifier alg_id;
   SecureVector<byte> pkcs8_key;

   DataSource_Memory key_source(PKCS8_PEM_handler(in, pass, encoding));
   PKCS8_extract_info(key_source, alg_id, pkcs8_key);

   const std::string alg_name = OIDS::lookup(alg_id.oid);
   if(alg_name == "")
      throw Decoding_Error("Unknown algorithm OID: " + alg_id.oid.as_string());

   std::auto_ptr<PKCS8_PrivateKey> key(get_private_key(alg_name));

   if(!key.get())
      throw Decoding_Error("Unknown PK algorithm/OID: " + alg_name + ", " +
                           alg_id.oid.as_string());

   Pipe output;
   output.process_msg(alg_id.parameters);
   output.process_msg(pkcs8_key);
   key->BER_decode_params(output);
   output.set_default_msg(1);
   key->BER_decode_priv(output);

   return key.release();
   }

}

}
