/*************************************************
* Global RNG Seeding Source File                 *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/rng.h>
#include <botan/conf.h>
#include <botan/es_file.h>

#if defined(BOTAN_EXT_ENTROPY_SRC_EGD)
  #include <botan/es_egd.h>
#endif

#if defined(BOTAN_EXT_ENTROPY_SRC_UNIX)
  #include <botan/es_unix.h>
#endif

#if defined(BOTAN_EXT_ENTROPY_SRC_FTW)
  #include <botan/es_ftw.h>
#endif

#if defined(BOTAN_EXT_ENTROPY_SRC_PTHREAD)
  #include <botan/es_pthr.h>
#endif

#if defined(BOTAN_EXT_ENTROPY_SRC_BEOS)
  #include <botan/es_beos.h>
#endif

namespace Botan {

namespace {

/*************************************************
* Try to do a poll on an EntropySource           *
*************************************************/
u32bit poll_es(EntropySource& es, bool slow_poll, u32bit buf_size = 256)
   {
   SecureVector<byte> buffer(buf_size);
   u32bit got = 0;

   if(slow_poll)
      got = es.slow_poll(buffer, buffer.size());
   else
      got = es.fast_poll(buffer, buffer.size());

   Global_RNG::add_entropy(buffer, got);
   return entropy_estimate(buffer, got);
   }

}

namespace Global_RNG {

/*************************************************
* Seed the global RNG                            *
*************************************************/
bool seed(bool slow_poll)
   {
   const u32bit min_entropy = Config::get_u32bit("rng/min_entropy");
   u32bit bits = 0;

   File_EntropySource es_file;
   bits += poll_es(es_file, slow_poll, 512);

#if defined(BOTAN_EXT_ENTROPY_SRC_EGD)
   EGD_EntropySource es_egd(Config::get_list("rng/egd_path"));
   bits += poll_es(es_egd, slow_poll, 128);
#endif

   if(bits >= min_entropy) return true;

#if defined(BOTAN_EXT_ENTROPY_SRC_BEOS)
   BeOS_EntropySource es_beos;
   bits += poll_es(es_beos, slow_poll);
#endif

#if defined(BOTAN_EXT_ENTROPY_SRC_UNIX)
   Unix_EntropySource es_unix;
   bits += poll_es(es_unix, slow_poll);
#endif

   if(bits >= min_entropy) return true;

#if defined(BOTAN_EXT_ENTROPY_SRC_FTW)
   FTW_EntropySource es_ftw;
   bits += poll_es(es_ftw, slow_poll);
#endif

   if(bits >= min_entropy) return true;

#if defined(BOTAN_EXT_ENTROPY_SRC_PTHREAD)
   Pthread_EntropySource es_pthr;
   bits += poll_es(es_pthr, slow_poll);
#endif

   return (bits >= min_entropy);
   }

}

}
