/*************************************************
* X.509 Certificate Authority Source File        *
* (C) 1999-2003 The Botan Project                *
*************************************************/

#include <botan/x509_ca.h>
#include <botan/x509stor.h>
#include <botan/conf.h>
#include <botan/lookup.h>
#include <botan/look_pk.h>
#include <botan/numthry.h>
#include <botan/oids.h>
#include <botan/rng.h>
#include <typeinfo>
#include <memory>

namespace Botan {

namespace {

/*************************************************
* Make sure the private key can sign             *
*************************************************/
void check_key(const PKCS8_PrivateKey& priv_key)
   {
   const PKCS8_PrivateKey* key = &priv_key;
   if(!dynamic_cast<const PK_Signing_Key*>(key))
      throw Invalid_Argument("X509_CA: " + key->algo_name() + " cannot sign");
   }

/*************************************************
* Find the allowable key constraints             *
*************************************************/
Key_Constraints find_constraints(const X509_PublicKey& pub_key,
                                 Key_Constraints limits)
   {
   const X509_PublicKey* key = &pub_key;
   u32bit constraints = 0;

   if(dynamic_cast<const PK_Encrypting_Key*>(key))
      constraints |= KEY_ENCIPHERMENT;

   if(dynamic_cast<const PK_Key_Agreement_Key*>(key))
      constraints |= KEY_AGREEMENT;

   if(dynamic_cast<const PK_Verifying_wo_MR_Key*>(key) ||
      dynamic_cast<const PK_Verifying_with_MR_Key*>(key))
      constraints |= DIGITAL_SIGNATURE | NON_REPUDATION;

   if(limits)
      constraints &= limits;

   return Key_Constraints(constraints);
   }

}

/*************************************************
* Load the keys and public certificate           *
*************************************************/
X509_CA::X509_CA(const X509_Certificate& c,
                 const PKCS8_PrivateKey& key) : cert(c)
   {
   check_key(key);
   if(!cert.is_ca_cert())
      throw Invalid_Argument("X509_CA: This certificate is not for a CA");

   std::string padding;
   Signature_Format format;

   Config::choose_format(key.algo_name(), padding, format);

   ca_sig_algo.oid = OIDS::lookup(key.algo_name() + "/" + padding);
   ca_sig_algo.parameters = key.DER_encode_params();

   const PK_Signing_Key& sig_key = dynamic_cast<const PK_Signing_Key&>(key);
   signer = get_pk_signer(sig_key, padding, format);
   }

/*************************************************
* Create a new root CA certificate               *
*************************************************/
X509_CA::X509_CA(const PKCS8_PrivateKey& key, X509_Cert_Options& opts)
   {
   check_key(key);

   std::string padding;
   Signature_Format format;

   Config::choose_format(key.algo_name(), padding, format);

   opts.CA_key();
   opts.self_signed = true;
   opts.set_constraints(Key_Constraints(KEY_CERT_SIGN | CRL_SIGN));

   ca_sig_algo.oid = OIDS::lookup(key.algo_name() + "/" + padding);
   ca_sig_algo.parameters = key.DER_encode_params();

   const PK_Signing_Key& sig_key = dynamic_cast<const PK_Signing_Key&>(key);
   signer = get_pk_signer(sig_key, padding, format);
   cert = sign_key(key, opts);
   }

/*************************************************
* Create a new certificate                       *
*************************************************/
X509_Certificate X509_CA::sign_key(const X509_PublicKey& pub_key,
                                   const X509_Cert_Options& opts) const
   {
   Pipe key_encoder;
   key_encoder.start_msg();
   X509::encode(pub_key, key_encoder, RAW_BER);
   key_encoder.end_msg();

   Key_Constraints constraints;

   if(opts.is_ca)
      {
      constraints = find_constraints(pub_key, NO_CONSTRAINTS);
      if((constraints & DIGITAL_SIGNATURE) == 0)
         throw Invalid_Argument("X509_CA: Cannot use this key for a CA");
      constraints = Key_Constraints(KEY_CERT_SIGN | CRL_SIGN);
      }
   else
      constraints = find_constraints(pub_key, opts.constraints);

   return make_cert(key_encoder.read_all(), opts, constraints);
   }

/*************************************************
* Create a new certificate                       *
*************************************************/
X509_Certificate X509_CA::make_cert(const SecureVector<byte>& pub_key,
                                    const X509_Cert_Options& opts,
                                    Key_Constraints constraints) const
   {
   opts.sanity_check();

   const u32bit X509_CERT_VERSION = 2;
   const u32bit SERIAL_BITS = 80;

   X509_DN subject_dn;
   subject_dn.add_attribute("X520.CommonName", opts.name);
   subject_dn.add_attribute("X520.Country", opts.country);
   subject_dn.add_attribute("X520.State", opts.state);
   subject_dn.add_attribute("X520.Locality", opts.locality);
   subject_dn.add_attribute("X520.Organization", opts.org);
   subject_dn.add_attribute("X520.OrganizationalUnit", opts.org_unit);

   X509_DN issuer_dn;

   if(opts.self_signed)
      issuer_dn = subject_dn;
   else
      issuer_dn = cert.subject_dn();

   DER_Encoder tbs_cert;

   tbs_cert.start_sequence();
   tbs_cert.start_explicit(ASN1_Tag(0));
   DER::encode(tbs_cert, X509_CERT_VERSION);
   tbs_cert.end_explicit(ASN1_Tag(0));

   DER::encode(tbs_cert, random_integer(SERIAL_BITS, Nonce));
   DER::encode(tbs_cert, ca_sig_algo);
   DER::encode(tbs_cert, issuer_dn);
   tbs_cert.start_sequence();
   DER::encode(tbs_cert, opts.start);
   DER::encode(tbs_cert, opts.end);
   tbs_cert.end_sequence();
   DER::encode(tbs_cert, subject_dn);
   tbs_cert.add_raw_octets(pub_key);

   tbs_cert.start_explicit(ASN1_Tag(3));
   tbs_cert.start_sequence();

   DER_Encoder v3_ext;

   std::auto_ptr<HashFunction> hash(get_hash("SHA-1"));
   hash->update(pub_key, pub_key.size());

   DER::encode(v3_ext, hash->final(), OCTET_STRING);
   do_ext(tbs_cert, v3_ext, "X509v3.SubjectKeyIdentifier", "subject_key_id");

   if(cert.subject_key_id().size())
      {
      v3_ext.start_sequence();
      v3_ext.start_explicit(ASN1_Tag(0));
      DER::encode(v3_ext, cert.subject_key_id(), OCTET_STRING);
      v3_ext.end_explicit(ASN1_Tag(0));
      v3_ext.end_sequence();
      do_ext(tbs_cert, v3_ext, "X509v3.AuthorityKeyIdentifier",
             "authority_key_id");
      }

   if(opts.is_ca ||
      (Config::get_string("x509/ca/basic_constraints") == "always"))
      {
      u32bit max_path_len = opts.path_limit;

      if(opts.is_ca && !opts.self_signed)
         {
         if(cert.path_len_constraint() == 0)
            throw Encoding_Error("X509_CA: This CA cannot sign new CA keys");

         max_path_len = std::min(max_path_len, cert.path_len_constraint() - 1);
         }

      v3_ext.start_sequence();
      if(opts.is_ca)
         {
         DER::encode(v3_ext, true);
         if(max_path_len != NO_CERT_PATH_LIMIT)
            DER::encode(v3_ext, max_path_len);
         }
      v3_ext.end_sequence();
      do_ext(tbs_cert, v3_ext, "X509v3.BasicConstraints", "basic_constraints");
      }

   AlternativeName subject_alt(opts.email, opts.uri, opts.dns);
   if(subject_alt.has_items())
      {
      DER::encode(v3_ext, subject_alt);
      do_ext(tbs_cert, v3_ext, "X509v3.SubjectAlternativeName",
             "subject_alternative_name");
      }

   if(constraints)
      {
      DER::encode(v3_ext, constraints);
      do_ext(tbs_cert, v3_ext, "X509v3.KeyUsage", "key_usage");
      }

   if(opts.ex_constraints.size())
      {
      v3_ext.start_sequence();
      for(u32bit j = 0; j != opts.ex_constraints.size(); j++)
         DER::encode(v3_ext, opts.ex_constraints[j]);
      v3_ext.end_sequence();
      do_ext(tbs_cert, v3_ext, "X509v3.ExtendedKeyUsage",
             "extended_key_usage");
      }

   tbs_cert.end_sequence();
   tbs_cert.end_explicit(ASN1_Tag(3));
   tbs_cert.end_sequence();

   SecureVector<byte> tbs_bits = tbs_cert.get_contents();
   SecureVector<byte> sig = signer->sign_message(tbs_bits);

   DER_Encoder full_cert;
   full_cert.start_sequence();
   full_cert.add_raw_octets(tbs_bits);
   DER::encode(full_cert, ca_sig_algo);
   DER::encode(full_cert, sig, BIT_STRING);
   full_cert.end_sequence();

   DataSource_Memory source(full_cert.get_contents());

   return X509_Certificate(source, RAW_BER);
   }

/*************************************************
* Handle encoding a v3 extension                 *
*************************************************/
void X509_CA::do_ext(DER_Encoder& new_cert, DER_Encoder& extension,
                     const std::string& oid, const std::string& opt) const
   {
   const std::string EXT_SETTING = Config::get_string("x509/exts/" + opt);

   if(EXT_SETTING == "")
      throw Exception("X509_CA: No policy setting for using " + oid);

   SecureVector<byte> ext_val = extension.get_contents();

   if(EXT_SETTING == "no")
      return;
   else if(EXT_SETTING == "yes" || EXT_SETTING == "noncritical" ||
           EXT_SETTING == "critical")
      {
      const bool critical = (EXT_SETTING == "critical");
      new_cert.start_sequence();
      DER::encode(new_cert, OIDS::lookup(oid));
      if(critical)
         DER::encode(new_cert, critical);
      DER::encode(new_cert, ext_val, OCTET_STRING);
      new_cert.end_sequence();
      }
   else
      throw Invalid_Argument("X509_CA:: Invalid value for option x509/exts/" +
                             opt + " of " + EXT_SETTING);
   }

/*************************************************
* Create a new, empty CRL                        *
*************************************************/
X509_CRL X509_CA::new_crl() const
   {
   std::vector<CRL_Entry> empty;
   return make_crl(empty, 1);
   }

/*************************************************
* Update a CRL with new entries                  *
*************************************************/
X509_CRL X509_CA::update_crl(const std::vector<CRL_Entry>& new_revoked,
                             const X509_CRL& crl) const
   {
   std::vector<CRL_Entry> full_revoked;
   std::vector<CRL_Entry> already_revoked = crl.get_revoked();

   X509_Store store;
   store.add_cert(cert, true);
   if(store.add_crl(crl) != VERIFIED)
      throw Invalid_Argument("X509_CA::update_crl: Invalid CRL provided");

   for(u32bit j = 0; j != already_revoked.size(); j++)
      full_revoked.push_back(already_revoked[j]);
   for(u32bit j = 0; j != new_revoked.size(); j++)
      full_revoked.push_back(new_revoked[j]);

   std::vector<CRL_Entry> cert_list;
   std::sort(full_revoked.begin(), full_revoked.end());
   std::unique_copy(full_revoked.begin(), full_revoked.end(),
                    std::back_inserter(cert_list));

   return make_crl(cert_list, crl.crl_number() + 1);
   }

/*************************************************
* Create a CRL                                   *
*************************************************/
X509_CRL X509_CA::make_crl(const std::vector<CRL_Entry>& revoked,
                           u32bit crl_number) const
   {
   const u32bit X509_CRL_VERSION = 1;
   const u32bit NEXT_UPDATE = Config::get_time("x509/crl/next_update");

   DER_Encoder tbs_crl;

   tbs_crl.start_sequence();
   DER::encode(tbs_crl, X509_CRL_VERSION);
   DER::encode(tbs_crl, ca_sig_algo);
   DER::encode(tbs_crl, cert.subject_dn());
   DER::encode(tbs_crl, X509_Time(system_time()));
   DER::encode(tbs_crl, X509_Time(system_time() + NEXT_UPDATE));

   if(revoked.size())
      {
      tbs_crl.start_sequence();
      for(u32bit j = 0; j != revoked.size(); j++)
         DER::encode(tbs_crl, revoked[j]);
      tbs_crl.end_sequence();
      }

   tbs_crl.start_explicit(ASN1_Tag(0));
   tbs_crl.start_sequence();

   DER_Encoder crl_ext;

   if(cert.subject_key_id().size())
      {
      crl_ext.start_sequence();
      crl_ext.start_explicit(ASN1_Tag(0));
      DER::encode(crl_ext, cert.subject_key_id(), OCTET_STRING);
      crl_ext.end_explicit(ASN1_Tag(0));
      crl_ext.end_sequence();
      do_ext(tbs_crl, crl_ext, "X509v3.AuthorityKeyIdentifier",
             "authority_key_id");
      }

   if(crl_number)
      {
      DER::encode(crl_ext, crl_number);
      do_ext(tbs_crl, crl_ext, "X509v3.CRLNumber", "crl_number");
      }

   tbs_crl.end_sequence();
   tbs_crl.end_explicit(ASN1_Tag(0));
   tbs_crl.end_sequence();

   SecureVector<byte> tbs_bits = tbs_crl.get_contents();
   SecureVector<byte> sig = signer->sign_message(tbs_bits);

   DER_Encoder full_crl;
   full_crl.start_sequence();
   full_crl.add_raw_octets(tbs_bits);
   DER::encode(full_crl, ca_sig_algo);
   DER::encode(full_crl, sig, BIT_STRING);
   full_crl.end_sequence();

   DataSource_Memory source(full_crl.get_contents());

   return X509_CRL(source, RAW_BER);
   }

/*************************************************
* Return the CA's certificate                    *
*************************************************/
X509_Certificate X509_CA::ca_certificate() const
   {
   return cert;
   }

/*************************************************
* X509_CA Destructor                             *
*************************************************/
X509_CA::~X509_CA()
   {
   delete signer;
   }

}
